#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc grammar specification.
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si: 

from parsimonious.grammar import Grammar

# Grammar according to http://project-haystack.org/doc/Zinc
zinc_grammar = Grammar(r'''
        grid        =   gridMeta cols row*
        nl          =   "\n" / "\r\n"
        gridMeta    =   ver meta? nl
        ver         =   "ver:" str
        meta        =   ( " " metaItem )+
        metaItem    =   metaPair / metaMarker
        metaMarker  =   id
        metaPair    =   id ":" " "* scalar
        valueSep    =   " "* "," " "*
        cols        =   col ( valueSep col )* nl
        col         =   id meta?
        row         =   cell ( valueSep cell )* nl
        cell        =   scalar?
        id          =   alphaLo ( alphaLo / alphaHi / digit / "_" )*
        scalar      =   ref / bin / str / uri / dateTime / date / time / coord / number / null / marker / remove / bool
        null        =   "N"
        marker      =   "M"
        remove      =   "R"
        bool        =   "T" / "F"
        bin         =   "Bin(" binChar* ")"
        binChar     =   ~r"[\x20-\x27\x2a-\x7f]"
        ref         =   "@" refChar* ( " " str )?
        refChar     =   alpha / digit / "_" / ":" / "-" / "." / "~"
        str         =   "\"" strChar* "\""
        uri         =   "`" uriChar* "`"
        strChar     =   ~r"([^\x00-\x1f\\\"]|\\[bfnrt\\\"]|\\u[0-9a-fA-F]{4})"
        uriChar     =   ~r"([^\x00-\x1f\\`]|\\[bfnrt\\:/?#\[\]@&=;`]|\\u[0-9a-fA-F]{4})"
        number      =   quantity / decimal / "INF" / "-INF" / "NaN"
        decimal     =   "-"? digits ( "." digits )? exp?
        quantity    =   decimal unit
        plusMinus   =   "+" / "-"
        exp         =   ( "e" / "E" ) plusMinus? digits
        unit        =   unitChar+
        unitChar    =   ~r"[^\x00-\x20,]"
        dateTime    =   isoDateTime ( " "+ timeZoneName )?
        isoDateTime =   date dateSep time tzHHMMOffset?
        date        =   digit digit digit digit "-" digit digit "-" digit digit
        time        =   digit digit ":" digit digit ":" digit digit ( "." digit+ )?
        dateSep     =   "t" / "T"
        timeZoneName=   tzUTCOffset / tzName
        tzUTCGMT    =   "UTC" / "GMT"
        tzUTCOffset =   tzUTCGMT ( "0" / ( plusMinus digit+ ) )?
        tzName      =   alphaHi tzNameChar*
        tzNameChar  =   alpha / digit / "_" / "-"
        tzHHMMOffset=   ~r"([zZ]|[+\-]\d{2}:?\d{2})"
        coord       =   "C(" " "* coordDeg valueSep coordDeg " "* ")"
        coordDeg    =   "-"? digits? ( "." digits )?
        alphaLo     =   ~r"[a-z]"
        alphaHi     =   ~r"[A-Z]"
        alpha       =   alphaLo / alphaHi
        digit       =   ~r"[0-9]"
        digits      =   digit ( digit / "_" )*
''')



# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

import hszinc
import datetime
import pytz
import random
import string
import six

STR_CHARSET = string.ascii_letters + string.digits + '\n\r\t\f\b'

def gen_random_const():
    return random.choice([True,False,None,hszinc.MARKER,hszinc.REMOVE])

def gen_random_ref():
    # Generate a randomised reference.
    name = gen_random_str(charset=\
                string.ascii_letters + string.digits\
                + '_:-.~')
    if random.choice([True,False]):
        value = gen_random_str()
    else:
        value = None

    return hszinc.Ref(name, value)

def gen_random_bin():
    # Generate a randomized binary
    return hszinc.Bin(random.choice([
        'text/plain',
        'text/html',
        'text/zinc',
        'application/json',
        'application/octet-stream',
        'image/png',
        'image/jpeg',
    ]))

def gen_random_uri():
    return hszinc.Uri(gen_random_str())

def gen_random_str(min_length=1, max_length=20,charset=STR_CHARSET):
    # Generate a random 20-character string
    return ''.join([random.choice(charset) for c in range(0,
        random.randint(min_length, max_length))])

def gen_random_date():
    # This might generate an invalid date, we keep trying until we get one.
    while True:
        try:
            return datetime.date(random.randint(1,3000),
                                random.randint(1,12), random.randint(1,31))
        except ValueError:
            pass

def gen_random_time():
    return datetime.time(random.randint(0,23), random.randint(0,59),
                        random.randint(0,59), random.randint(0,999999))

def gen_random_date_time():
    # Pick a random timezone
    tz_name = random.choice(list(hszinc.zoneinfo.get_tz_map().keys()))
    tz = hszinc.zoneinfo.timezone(tz_name)
    return tz.localize(datetime.datetime.combine(
        gen_random_date(), gen_random_time()))

def gen_random_coordinate():
    return hszinc.Coordinate(\
            gen_random_num(360)-180.0,
            gen_random_num(360)-180.0)

def gen_random_num(scale=1000,digits=2):
    return round(random.random()*scale, digits)

def gen_random_quantity():
    return hszinc.Quantity(gen_random_num(),
            gen_random_str(max_length=8,
                charset=string.ascii_letters))

RANDOM_TYPES = [
    gen_random_const, gen_random_ref, gen_random_bin, gen_random_uri,
    gen_random_str, gen_random_date, gen_random_time, gen_random_date_time,
    gen_random_coordinate, gen_random_num, gen_random_quantity
]

def gen_random_scalar():
    return random.choice(RANDOM_TYPES)()

def gen_random_name(existing=None):
    while True:
        meta = random.choice(string.ascii_lowercase) \
                + gen_random_str(min_length=0, max_length=7, \
                    charset=string.ascii_letters + string.digits)
        if (existing is None) or (meta not in existing):
            return meta

def gen_random_meta():
    meta = hszinc.MetadataObject()
    names = set()
    for n in range(0, random.randint(1,5)):
        name = gen_random_name(existing=names)
        value = gen_random_scalar()
        meta[name] = value
    return meta

def dump_grid(g):
    print ('Metadata:')
    for k, v in g.metadata.items():
        print ('   %s = %r' % (k, v))
    print ('Columns:')
    for c, meta in g.column.items():
        print ('   %s:' % c)
        for k, v in g.column[c].items():
            print ('      %s = %r' % (k, v))
    print ('Rows:')
    for row in g:
        print ('---')
        for c, v in row.items():
            print ('  %s = %r' % (c, v))

def approx_check(v1, v2):
    # Check types match
    if not (isinstance(v1, six.string_types) \
            and isinstance(v2, six.string_types)):
        assert type(v1) == type(v2), '%s != %s' % (type(v1), type(v2))
    if isinstance(v1, datetime.time):
        assert v1.replace(microsecond=0) == v2.replace(microsecond=0)
        approx_check(v1.microsecond, v2.microsecond)
    elif isinstance(v1, datetime.datetime):
        assert v1.tzinfo == v2.tzinfo
        assert v1.date() == v2.date()
        approx_check(v1.time(), v2.time())
    elif isinstance(v1, hszinc.Quantity):
        assert v1.unit == v2.unit
        approx_check(v1.value, v2.value)
    elif isinstance(v1, hszinc.Coordinate):
        approx_check(v1.latitude, v2.latitude)
        approx_check(v1.longitude, v2.longitude)
    elif isinstance(v1, float):
        assert abs(v1 - v2) < 0.000001
    else:
        assert v1 == v2, '%r != %r' % (v1, v2)

def try_dump_parse():
    # Generate a randomised grid of values and try parsing it back.
    ref_grid = hszinc.Grid()
    ref_grid.metadata.extend(gen_random_meta())

    # Randomised columns
    for n in range(0, random.randint(1,5)):
        col_name = gen_random_name(existing=ref_grid.column)
        if random.choice([True,False]):
            ref_grid.column[col_name] = gen_random_meta()
        else:
            ref_grid.column[col_name] = {}

    # Randomised rows
    for n in range(0, random.randint(0,20)):
        row = {}
        for c in ref_grid.column.keys():
            if random.choice([True,False]):
                row[c] = gen_random_scalar()
        ref_grid.append(row)

    try:
        # Dump the randomised grid to a string
        grid_str = hszinc.dump(ref_grid)
    except:
        # Dump some detail about the grid
        print ('Failed to dump grid.')
        dump_grid(ref_grid)
        raise

    # Parse the grid string
    try:
        grid_list = hszinc.parse(grid_str)
    except:
        print ('Failed to parse dumped grid')
        dump_grid(ref_grid)
        print ('--- Parsed string ---')
        print (grid_str)
        raise

    assert len(grid_list) == 1
    parsed_grid = grid_list.pop(0)

    # Check metadata matches
    try:
        assert list(ref_grid.metadata.keys()) \
                == list(parsed_grid.metadata.keys())
        for key in ref_grid.metadata.keys():
            approx_check(ref_grid.metadata[key], parsed_grid.metadata[key])
    except:
        print ('Mismatch in metadata')
        print ('Reference grid')
        dump_grid(ref_grid)
        print ('Parsed grid')
        dump_grid(parsed_grid)
        raise

    try:
        # Check column matches
        assert list(ref_grid.column.keys()) \
                == list(parsed_grid.column.keys())
    except:
        print ('Mismatch in column')
        print ('Reference grid')
        dump_grid(ref_grid)
        print ('Parsed grid')
        dump_grid(parsed_grid)
        raise
    for col in ref_grid.column.keys():
        try:
            for key in ref_grid.column[col].keys():
                approx_check(ref_grid.column[col][key], \
                        parsed_grid.column[col][key])
        except:
            print ('Mismatch in metadata for column %s' % col)
            print ('Reference: %r' % ref_grid.column[col])
            print ('Parsed:    %r' % parsed_grid.column[col])
            raise

    try:
        # Check row matches
        assert len(ref_grid) == len(parsed_grid)
    except:
        print ('Mismatch in row count')
        print ('Reference grid')
        dump_grid(ref_grid)
        print ('Parsed grid')
        dump_grid(parsed_grid)

    for (ref_row, parsed_row) in zip(ref_grid, parsed_grid):
        try:
            for col in ref_grid.column.keys():
                approx_check(ref_row.get(col), parsed_row.get(col))
        except:
            print ('Mismatch in row')
            print ('Reference:')
            print (ref_row)
            print ('Parsed:')
            print (parsed_row)
            raise

def test_loopback():
    for trial in range(0, 10):
        try_dump_parse()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc Grid dumper
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from .grid import Grid
from .sortabledict import SortableDict
from .datatypes import Quantity, Coordinate, Ref, Bin, Uri, \
        MARKER, REMOVE, STR_SUB
from .zoneinfo import timezone_name
from .parser import MODE_ZINC, MODE_JSON, MARKER_STR
import datetime
import iso8601
import re
import functools
import json
import six

URI_META = re.compile(six.u(r'([\\`\u0080-\uffff])'))
STR_META = re.compile(six.u(r'([\\"\u0080-\uffff])'))

def str_sub(match):
    c = match.group(0)
    o = ord(c)
    if o >= 0x0080:
        # Unicode
        return '\\u%04x' % o
    elif c in '\\"':
        return '\\%s' % c

def uri_sub(match):
    c = match.group(0)
    o = ord(c)
    if o >= 0x80:
        # Unicode
        return '\\u%04x' % o
    elif c in '\\`':
        return '\\%s' % c

def dump(grids, mode=MODE_ZINC):
    '''
    Parse the given Zinc text and return the equivalent data.
    '''
    if isinstance(grids, Grid):
        return dump_grid(grids, mode=mode)
    _dump = functools.partial(dump_grid, mode=mode)
    if mode == MODE_ZINC:
        return '\n'.join(map(_dump, grids))
    elif mode == MODE_JSON:
        return '[%s]' % ','.join(map(_dump, grids))
    else: # pragma: no cover
        raise NotImplementedError('Format not implemented: %s' % mode)

def dump_grid(grid, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        header = 'ver:%s' % dump_str(grid._version)
        if bool(grid.metadata):
            header += ' ' + dump_meta(grid.metadata)
        columns = dump_columns(grid.column)
        rows = dump_rows(grid)
        return '\n'.join([header, columns] + rows + [''])
    elif mode == MODE_JSON:
        return json.dumps({
            'meta': dump_meta(grid.metadata, version=grid._version,
                mode=mode),
            'cols': dump_columns(grid.column, mode=mode),
            'rows': dump_rows(grid, mode=mode),
        })
    else: # pragma: no cover
        raise NotImplementedError('Format not implemented: %s' % mode)

def dump_meta(meta, version=None, mode=MODE_ZINC):
    _dump = functools.partial(dump_meta_item, mode=mode)
    if mode == MODE_ZINC:
        return ' '.join(map(_dump, list(meta.items())))
    else:
        _meta = dict(map(_dump, list(meta.items())))
        if version is not None:
            _meta['ver'] = version
        return _meta

def dump_meta_item(item, mode=MODE_ZINC):
    (item_id, item_value) = item
    if mode == MODE_ZINC:
        if item_value is MARKER:
            return dump_id(item_id, mode=mode)
        else:
            return '%s:%s' % (dump_id(item_id, mode=mode), \
                    dump_scalar(item_value, mode=mode))
    elif mode == MODE_JSON:
        return (dump_id(item_id, mode=mode), \
                dump_scalar(item_value, mode=mode))

def dump_columns(cols, mode=MODE_ZINC):
    _dump = functools.partial(dump_column, mode=mode)
    _cols = list(zip(*list(cols.items())))
    if mode == MODE_ZINC:
        return ','.join(map(_dump, *_cols))
    elif mode == MODE_JSON:
        return list(map(_dump, *_cols))

def dump_column(col, col_meta, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        if bool(col_meta):
            return '%s %s' % (dump_id(col, mode=mode), \
                    dump_meta(col_meta, mode=mode))
        else:
            return dump_id(col, mode=mode)
    elif mode == MODE_JSON:
        if bool(col_meta):
            _meta = dump_meta(col_meta, mode=mode)
        else:
            _meta = {}
        _meta['name'] = col
        return _meta

def dump_rows(grid, mode=MODE_ZINC):
    return [dump_row(grid, r, mode=mode) for r in grid]

def dump_row(grid, row, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        return ','.join([dump_scalar(row.get(c), mode=mode) for \
                c in list(grid.column.keys())])
    elif mode == MODE_JSON:
        return dict([
            (c, dump_scalar(row.get(c), mode=mode))
            for c in list(grid.column.keys())])

def dump_scalar(scalar, mode=MODE_ZINC):
    if scalar is None:
        if mode == MODE_ZINC:
            return 'N'
        elif mode == MODE_JSON:
            return None
        else: # pragma: no cover
            raise NotImplementedError('Unhandled case: %r' % scalar)
    elif scalar is MARKER:
        if mode == MODE_ZINC:
            return 'M'
        elif mode == MODE_JSON:
            return MARKER_STR
        else: # pragma: no cover
            raise NotImplementedError('Unhandled case: %r' % scalar)
    elif scalar is REMOVE:
        if mode == MODE_ZINC:
            return 'R'
        else: # pragma: no cover
            raise NotImplementedError('Unhandled case: %r' % scalar)
    elif isinstance(scalar, bool):
        return dump_bool(scalar, mode=mode)
    elif isinstance(scalar, Ref):
        return dump_ref(scalar, mode=mode)
    elif isinstance(scalar, Bin):
        return dump_bin(scalar, mode=mode)
    elif isinstance(scalar, Uri):
        return dump_uri(scalar, mode=mode)
    elif isinstance(scalar, six.string_types):
        return dump_str(scalar, mode=mode)
    elif isinstance(scalar, datetime.datetime):
        return dump_date_time(scalar, mode=mode)
    elif isinstance(scalar, datetime.time):
        return dump_time(scalar, mode=mode)
    elif isinstance(scalar, datetime.date):
        return dump_date(scalar, mode=mode)
    elif isinstance(scalar, Coordinate):
        return dump_coord(scalar, mode=mode)
    elif isinstance(scalar, Quantity):
        return dump_quantity(scalar, mode=mode)
    elif isinstance(scalar, float) or \
            isinstance(scalar, int) or \
            isinstance(scalar, int):
        return dump_decimal(scalar, mode=mode)
    else: # pragma: no cover
        raise NotImplementedError('Unhandled case: %r' % scalar)

def dump_id(id_str, mode=MODE_ZINC):
    return id_str

def dump_str(str_value, mode=MODE_ZINC):
    if mode == MODE_JSON:
        return u's:%s' % str_value
    elif mode == MODE_ZINC:
        # Replace special characters.
        str_value = STR_META.sub(str_sub, str_value)
        # Replace other escapes.
        for orig, esc in STR_SUB:
            str_value = str_value.replace(orig, esc)
        return '"%s"' % str_value

def dump_uri(uri_value, mode=MODE_ZINC):
    if mode == MODE_JSON:
        return u'u:%s' % uri_value
    elif mode == MODE_ZINC:
        # Replace special characters.
        uri_value = URI_META.sub(uri_sub, uri_value)
        # Replace other escapes.
        for orig, esc in STR_SUB:
            uri_value = uri_value.replace(orig, esc)
        return '`%s`' % uri_value

def dump_bin(bin_value, mode=MODE_ZINC):
    if mode == MODE_JSON:
        return u'b:%s' % bin_value
    elif mode == MODE_ZINC:
        return 'Bin(%s)' % bin_value

def dump_quantity(quantity, mode=MODE_ZINC):
    if quantity.unit is None:
        return dump_decimal(quantity.value, mode=mode)
    elif mode == MODE_ZINC:
        return '%s%s' % (dump_decimal(quantity.value), quantity.unit)
    elif mode == MODE_JSON:
        return 'n:%f %s' % (quantity.value, quantity.unit)

def dump_decimal(decimal, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        return str(decimal)
    elif mode == MODE_JSON:
        return 'n:%f' % decimal

def dump_bool(bool_value, mode=MODE_ZINC):
    if mode == MODE_JSON:
        return bool_value
    elif mode == MODE_ZINC:
        return 'T' if bool(bool_value) else 'F'

def dump_coord(coordinate, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        return 'C(%f,%f)' % (coordinate.latitude, coordinate.longitude)
    elif mode == MODE_JSON:
        return 'c:%f,%f' % (coordinate.latitude, coordinate.longitude)

def dump_ref(ref, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        if ref.has_value:
            return '@%s %s' % (ref.name, dump_str(ref.value))
        else:
            return '@%s' % ref.name
    elif mode == MODE_JSON:
        if ref.has_value:
            return u'r:%s %s' % (ref.name, ref.value)
        else:
            return u'r:%s' % ref.name

def dump_date(date, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        return date.isoformat()
    elif mode == MODE_JSON:
        return 'd:%s' % date.isoformat()

def dump_time(time, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        return time.isoformat()
    elif mode == MODE_JSON:
        return 'h:%s' % time.isoformat()

def dump_date_time(date_time, mode=MODE_ZINC):
    tz_name = timezone_name(date_time)
    if mode == MODE_ZINC:
        return '%s %s' % (date_time.isoformat(), tz_name)
    elif mode == MODE_JSON:
        return 't:%s %s' % (date_time.isoformat(), tz_name)



# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

import hszinc
from copy import copy, deepcopy

def check_singleton_deepcopy(S):
    orig_dict = {'some_value': S}
    copy_dict = deepcopy(orig_dict)
    assert copy_dict['some_value'] is S

def test_marker_deepcopy():
    check_singleton_deepcopy(hszinc.MARKER)

def test_remove_deepcopy():
    check_singleton_deepcopy(hszinc.REMOVE)

def check_singleton_copy(S):
    assert copy(S) is S

def test_marker_copy():
    check_singleton_copy(hszinc.MARKER)

def test_remove_copy():
    check_singleton_copy(hszinc.REMOVE)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc Grid
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from .metadata import MetadataObject
from .sortabledict import SortableDict
from collections import MutableSequence

class Grid(MutableSequence):
    '''
    A grid is basically a series of tabular records.  The grid has a header
    which describes some metadata about the grid and its columns.  This is
    followed by zero or more rows.
    '''

    # Grid version number
    DEFAULT_VERSION = "2.0"

    def __init__(self, version=DEFAULT_VERSION, metadata=None, columns=None):
        '''
        Create a new Grid.
        '''
        # Version
        self._version   = version

        # Metadata
        self.metadata   = MetadataObject()

        # The columns
        self.column     = SortableDict()

        # Rows
        self._row       = []

        if metadata is not None:
            self.metadata.update(metadata.items())

        if columns is not None:
            if isinstance(columns, dict) or isinstance(columns, SortableDict):
                columns = list(columns.items())

            for col_id, col_meta in columns:
                if not isinstance(col_meta, MetadataObject):
                    mo = MetadataObject()
                    mo.extend(col_meta)
                    col_meta = mo
                self.column.add_item(col_id, col_meta)

    def __repr__(self):
        '''
        Return a representation of this grid.
        '''
        parts = []
        if bool(self.metadata):
            parts.append(u'\tMetadata: %s' % self.metadata)

        column_meta = []
        for col, col_meta in self.column.items():
            if bool(col_meta):
                column_meta.append(u'\t\t%s: %s' % (col, col_meta))
            else:
                column_meta.append(u'\t\t%s' % col)

        if bool(column_meta):
            parts.append(u'\tColumns:\n%s' % '\n'.join(column_meta))
        elif len(self.column):
            parts.append(u'\tColumns: %s' % ', '.join(self.column.keys()))
        else:
            parts.append(u'\tNo columns')

        if bool(self):
            parts.extend([
                u'\tRow %4d: %s' % (row, u', '.join([
                    ((u'%s=%r' % (col, data[col])) \
                            if col in data else \
                    (u'%s absent' % col)) for col \
                    in self.column.keys()]))
                for (row, data) in enumerate(self)
            ])
        else:
            parts.append(u'\tNo rows')
        # Represent as pseudo-XML
        class_name = self.__class__.__name__
        return u'<%s>\n%s\n</%s>' % (
                class_name, u'\n'.join(parts), class_name
        )

    def __getitem__(self, index):
        '''
        Retrieve the row at index.
        '''
        return self._row[index]

    def __len__(self):
        '''
        Return the number of rows in the grid.
        '''
        return len(self._row)

    def __setitem__(self, index, value):
        '''
        Replace the row at index.
        '''
        if not isinstance(value, dict):
            raise TypeError('value must be a dict')
        self._row[index] = value

    def __delitem__(self, index):
        '''
        Delete the row at index.
        '''
        del self._row[index]

    def insert(self, index, value):
        '''
        Insert a new row before index.
        '''
        if not isinstance(value, dict):
            raise TypeError('value must be a dict')
        self._row.insert(index, value)



# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

try:
    from .grid import Grid
    from .dumper import dump, dump_scalar
    from .parser import parse, MODE_JSON, MODE_ZINC
    from .metadata import MetadataObject
    from .datatypes import Quantity, Coordinate, Uri, Bin, MARKER, REMOVE, Ref

    __all__ = ['Grid', 'dump', 'parse', 'MetadataObject', 'Quantity',
            'Coordinate', 'Uri', 'Bin', 'MARKER', 'REMOVE', 'Ref',
            'MODE_JSON', 'MODE_ZINC']
except ImportError: # pragma: no cover
    # For setup.py to interrogate the version information.  This should *NOT*
    # get executed in production, and if it did, things wouldn't work anyway.
    pass

__author__ = 'VRT Systems'
__copyright__ = 'Copyright 2016, VRT Systems'
__credits__ = ['VRT Systems']
__license__ = 'BSD'
__version__ = '0.0.8'
__maintainer__ = 'VRT Systems'
__email__ = 'support@vrt.com.au'



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc Grid Metadata
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from .grid import Grid
from .metadata import MetadataObject
from .grammar import zinc_grammar
from .sortabledict import SortableDict
from .datatypes import Quantity, Coordinate, Ref, Bin, Uri, \
        MARKER, REMOVE, STR_SUB
from .zoneinfo import timezone
import datetime
import iso8601
import re
import six
import functools
import json
import copy

URI_META = re.compile(r'\\([:/\?#\[\]@\\&=;"$`])')
GRID_SEP = re.compile(r'\n\n+')

MODE_ZINC = 'zinc'
MODE_JSON = 'json'

# Type regular expressions
MARKER_STR  = 'm:'
NUMBER_RE   = re.compile(r'^n:(-?\d+(:?\.\d+)?(:?[eE][+\-]?\d+)?)(:? (.*))?$')
REF_RE      = re.compile(r'^r:([a-zA-Z0-9_:\-.~]+)(:? (.*))?$')
STR_RE      = re.compile(r'^s:(.*)$')
DATE_RE     = re.compile(r'^d:(\d{4})-(\d{2})-(\d{2})$')
TIME_RE     = re.compile(r'^h:(\d{2}):(\d{2})(:?:(\d{2}(:?\.\d+)?))?$')
DATETIME_RE = re.compile(r'^t:(\d{4}-\d{2}-\d{2}T'\
        r'\d{2}:\d{2}(:?:\d{2}(:?\.\d+)?)'\
        r'(:?[zZ]|[+\-]\d+:?\d*))(:? ([A-Za-z\-+_0-9]+))?$')
URI_RE      = re.compile(r'u:(.+)$')
BIN_RE      = re.compile(r'b:(.+)$')
COORD_RE    = re.compile(r'c:(-?\d*\.?\d*),(-?\d*\.?\d*)$')

STR_ESC_RE  = re.compile(r'\\([bfnrt"\\$]|u[0-9a-fA-F]{4})')

def parse(grid_str, mode=MODE_ZINC):
    '''
    Parse the given Zinc text and return the equivalent data.
    '''
    # Split the separate grids up, the grammar definition has trouble splitting
    # them up normally.  This will truncate the newline off the end of the last
    # row.
    _parse = functools.partial(parse_grid, mode=mode)
    if mode == MODE_JSON:
        if isinstance(grid_str, six.string_types):
            grid_data = json.loads(grid_str)
        else:
            grid_data = grid_str
        if isinstance(grid_data, dict):
            return _parse(grid_data)
        else:
            return list(map(_parse, grid_data))
    else:
        return list(map(_parse, GRID_SEP.split(grid_str.rstrip())))

def parse_grid(grid_str, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        if not grid_str.endswith('\n'):
            grid_str += '\n'

        parsed = zinc_grammar.parse(grid_str)
        assert len(parsed.children) == 3

        # First child is metadata
        meta = parsed.children[0]
        assert meta.expr_name == 'gridMeta'

        # We expect the version string next
        assert meta.children[0].expr_name == 'ver'
        version = parse_str(meta.children[0].children[1])

        # The metadata should be in a nameless node.  Parse the children.
        if bool(meta.children[1].children):
            meta_items = parse_meta(meta.children[1].children[0])
        else:
            # No metadata
            meta_items = MetadataObject()

        # The last thing in the list should be a newline
        assert meta.children[2].expr_name == 'nl'

        # The next item should be the columns
        columns = parse_columns(parsed.children[1])

        # This is enough to construct the base grid.
        grid = Grid(version=version, metadata=meta_items, columns=columns)

        # Now to parse the rows
        if bool(columns):
            parse_rows(grid, parsed.children[2])
    elif mode == MODE_JSON:
        # Grab the metadata
        if isinstance(grid_str, six.string_types):
            parsed = json.loads(grid_str)
        else:
            parsed = copy.deepcopy(grid_str)
        meta = parsed.pop('meta')
        version = meta.pop('ver')

        # Parse the remaining elements
        metadata = {}
        for name, value in meta.items():
            metadata[name] = parse_scalar(value, mode=mode)

        grid = Grid(version=version, metadata=metadata)

        # Grab the columns in the order given
        for col in parsed.pop('cols'):
            name = col.pop('name')
            meta = {}
            for key, value in col.items():
                meta[key] = parse_scalar(value, mode=mode)
            grid.column[name] = meta

        # Parse the rows
        for row in (parsed.pop('rows',[]) or []):
            parsed_row = {}
            for col, value in row.items():
                parsed_row[col] = parse_scalar(value, mode=mode)
            grid.append(parsed_row)
    else: # pragma: no cover
        raise NotImplementedError('Format not implemented: %s' % mode)

    return grid

def parse_meta(meta):
    assert meta.expr_name == 'meta'
    items = MetadataObject()

    for item_parent in meta.children:
        # There'll probably be a space beforehand
        for node in item_parent.children:
            if node.expr_name == 'metaItem':
                (item_id, item_value) = parse_meta_item(node)
                items[item_id] = item_value
                break
    return items

def parse_meta_item(item):
    # This could be an 'id' or a 'metaPair'
    assert len(item.children) == 1
    item_child = item.children[0]
    if item_child.expr_name == 'id':
        # This is a marker
        return (parse_id(item_child), MARKER)
    elif item_child.expr_name == 'metaPair':
        # This is a metadata pair
        assert len(item_child.children) == 4
        item_id = parse_id(item_child.children[0])
        item_value = parse_scalar(item_child.children[-1])
        return (item_id, item_value)
    else: # pragma: no cover
        raise NotImplementedError('Unhandled case: %s' \
                % item_child.prettily())

def parse_columns(cols):
    assert cols.expr_name == 'cols'
    assert len(cols.children) == 3
    # The first column will be the first child, remaining will be wrapped in
    # a parent.
    (col_name, col) = parse_column(cols.children[0])
    columns = SortableDict()
    columns[col_name] = col

    if bool(cols.children[1]):
        for node in cols.children[1].children:
            for nchild in node.children:
                if nchild.expr_name == 'valueSep':
                    continue
                else:
                    (col_name, col) = parse_column(nchild)
                    columns[col_name] = col
    return columns

def parse_column(col):
    assert col.expr_name == 'col'
    # We should have the ID and the metadata
    assert len(col.children) == 2
    col_id = parse_id(col.children[0])
    if bool(col.children[1].children):
        col_meta = parse_meta(col.children[1].children[0])
    else:
        col_meta = MetadataObject()
    return (col_id, col_meta)

def parse_rows(grid, rows):
    for row in [r for r in rows.children if r.expr_name == 'row']:
        parse_row(grid, row)

def parse_row(grid, row):
    assert row.expr_name == 'row'
    # Cell names
    columns = list(grid.column.keys())
    if not bool(row.children): # pragma: no cover
        # Empty row?  Shouldn't happen, but just in case.
        return {}

    # First cell
    cell_id = columns.pop(0)
    cell_value = parse_cell(row.children[0])
    data = {cell_id: cell_value}

    # Remaining cells
    if bool(row.children[1].children) and bool(columns):
        for node in row.children[1].children:
            for nchild in node.children:
                if nchild.expr_name == 'valueSep':
                    continue
                try:
                    cell_id = columns.pop(0)
                except IndexError:
                    break

                cell_value = parse_cell(nchild)
                data[cell_id] = cell_value

    grid.append(data)

def parse_cell(cell):
    assert cell.expr_name == 'cell'
    if bool(cell.children):
        assert len(cell.children) == 1
        return parse_scalar(cell.children[0])
    else:
        return None

def parse_scalar(scalar, mode=MODE_ZINC, charset='utf-8'):
    if mode == MODE_ZINC:
        # Should have a child which is the type of scalar.
        assert scalar.expr_name == 'scalar'
        assert len(scalar.children) == 1
        scalar_child = scalar.children[0]
        if scalar_child.expr_name == 'null':
            return None
        elif scalar_child.expr_name == 'marker':
            return MARKER
        elif scalar_child.expr_name == 'remove':
            return REMOVE
        elif scalar_child.expr_name == 'bool':
            return parse_bool(scalar_child)
        elif scalar_child.expr_name == 'ref':
            return parse_ref(scalar_child, charset)
        elif scalar_child.expr_name == 'bin':
            return parse_bin(scalar_child)
        elif scalar_child.expr_name == 'str':
            return parse_str(scalar_child, charset)
        elif scalar_child.expr_name == 'uri':
            return parse_uri(scalar_child)
        elif scalar_child.expr_name == 'date':
            return parse_date(scalar_child)
        elif scalar_child.expr_name == 'time':
            return parse_time(scalar_child)
        elif scalar_child.expr_name == 'dateTime':
            return parse_date_time(scalar_child)
        elif scalar_child.expr_name == 'coord':
            return parse_coord(scalar_child)
        elif scalar_child.expr_name == 'number':
            return parse_number(scalar_child)
        else: # pragma: no cover
            raise NotImplementedError('Unhandled case: %s' \
                    % scalar_child.prettily())
    elif mode == MODE_JSON:
        # Simple cases
        if scalar is None:
            return None
        elif scalar == MARKER_STR:
            return MARKER
        elif isinstance(scalar, bool):
            return scalar
        elif scalar == 'n:INF':
            return float('INF')
        elif scalar == 'n:-INF':
            return -float('INF')
        elif scalar == 'n:NaN':
            return float('nan')
        # Conversion to dict of float value turn them into float 
        # so regex won't work... better just return them
        elif isinstance(scalar, float):
            return scalar

        # Is it a number?
        match = NUMBER_RE.match(scalar)
        if match:
            # We'll get a value and a unit, amongst other tokens.
            matched = match.groups()
            value = float(matched[0])
            if matched[-1] is not None:
                # It's a quantity
                return Quantity(value, matched[-1])
            else:
                # It's a raw value
                return value

        # Is it a string?
        match = STR_RE.match(scalar)
        if match:
            return match.group(1)

        # Is it a reference?
        match = REF_RE.match(scalar)
        if match:
            matched = match.groups()
            if matched[-1] is not None:
                return Ref(matched[0], matched[-1], has_value=True)
            else:
                return Ref(matched[0])

        # Is it a date?
        match = DATE_RE.match(scalar)
        if match:
            (year, month, day) = match.groups()
            return datetime.date(year=int(year), month=int(month), day=int(day))

        # Is it a time?
        match = TIME_RE.match(scalar)
        if match:
            (hour, minute, _, second, _) = match.groups()
            # Convert second to seconds and microseconds
            if second is None:
                sec = 0
                usec = 0
            elif '.' in second:
                (whole_sec, frac_sec) = second.split('.',1)
                sec = int(whole_sec)
                usec = int(frac_sec[:6].ljust(6,'0'))
            else:
                sec = int(second)
                usec = 0
            return datetime.time(hour=int(hour), minute=int(minute),
                    second=sec, microsecond=usec)

        # Is it a date/time?
        match = DATETIME_RE.match(scalar)
        if match:
            matches = match.groups()
            # Parse ISO8601 component
            isodate = iso8601.parse_date(matches[0])
            # Parse timezone
            tzname  = matches[-1]
            if tzname is None:
                return isodate  # No timezone given
            else:
                try:
                    tz = timezone(tzname)
                    return isodate.astimezone(tz)
                except: # pragma: no cover
                    # Unlikely code path.
                    return isodate

        # Is it a URI?
        match = URI_RE.match(scalar)
        if match:
            return Uri(match.group(1))

        # Is it a Bin?
        match = BIN_RE.match(scalar)
        if match:
            return Bin(match.group(1))

        # Is it a co-ordinate?
        match = COORD_RE.match(scalar)
        if match:
            (lat,lng) = match.groups()
            return Coordinate(float(lat),float(lng))
        return scalar
    else: # pragma: no cover
        raise NotImplementedError('Format not implemented: %s' % mode)

def parse_id(id_node):
    assert id_node.expr_name == 'id'
    return id_node.text

# These *should* have coverage, but unfortunately, there's no way to
# say, "no cover on Python 2.x" or "no cover on Python 3.x"
if six.PY3: # pragma: no cover
    # This will probably need investigation... using nHastack, 
    # we need to use utf-8 instead of ascii... 
    # charset should be defined somewhere
    _str_to_bytes = lambda s, charset : six.binary_type(s, charset)
else: # pragma: no cover
    _str_to_bytes = lambda s, charset : six.text_type(s)

def _decode_escape(esc):
    esc = esc.group(1)
    if esc[0] in 'uU':
        return six.unichr(int(esc[1:], 16))
    elif esc == '"':
        return '\"'
    elif esc == 'b':
        return '\b'
    elif esc == 'f':
        return '\f'
    elif esc == 'n':
        return '\n'
    elif esc == 'r':
        return '\r'
    elif esc == 't':
        return '\t'
    else:
        return esc

def _sub_escape(str_value):
    (str_value, replacements) = STR_ESC_RE.subn(_decode_escape, str_value)
    while replacements > 0:
        (str_value, replacements) = STR_ESC_RE.subn(_decode_escape, str_value)
    return str_value


def parse_str(str_node, charset = 'utf-8'):
    assert str_node.expr_name == 'str'
    assert len(str_node.children) == 3
    return _sub_escape(str_node.children[1].text)

def parse_uri(uri_node):
    assert uri_node.expr_name == 'uri'
    assert len(uri_node.children) == 3

    uri_value = uri_node.children[1].text

    # Replace escapes.
    for orig, esc in STR_SUB:
        uri_value = uri_value.replace(esc, orig)
    return Uri(URI_META.sub(\
            lambda m : m.group(0) if m.group(1) != '`' else '`',
            uri_value))

def parse_bin(bin_node):
    assert bin_node.expr_name == 'bin'
    assert len(bin_node.children) == 3
    return Bin(bin_node.children[1].text)

def parse_number(num_node):
    assert num_node.expr_name == 'number'
    assert len(num_node.children) == 1
    value = num_node.children[0]
    if value.expr_name == 'quantity':
        return parse_quantity(value)
    elif value.expr_name == 'decimal':
        return parse_decimal(value)
    elif value.text in ('INF','-INF','NaN'):
        return float(value.text)
    else: # pragma: no cover
        raise NotImplementedError(\
                'Unhandled case: %s' % value.prettily())

def parse_quantity(quantity_node):
    assert quantity_node.expr_name == 'quantity'
    assert len(quantity_node.children) == 2
    # First child is the value, type decimal
    value = parse_decimal(quantity_node.children[0])
    # Second is the unit.
    unit = quantity_node.children[1].text
    return Quantity(value, unit)

def parse_decimal(decimal_node):
    assert decimal_node.expr_name == 'decimal'
    return float(decimal_node.text.replace('_',''))

def parse_bool(bool_node):
    assert bool_node.expr_name == 'bool'
    assert bool_node.text.upper() in 'TF'
    return bool_node.text.upper() == 'T'

def parse_coord(coordinate_node):
    assert coordinate_node.expr_name == 'coord'
    assert len(coordinate_node.children) == 7
    # Co-ordinates are in child nodes 2 and 4.
    lat = float(coordinate_node.children[2].text)
    lng = float(coordinate_node.children[4].text)
    return Coordinate(lat, lng)

def parse_ref(ref_node, charset = 'utf-8'):
    assert ref_node.expr_name == 'ref'
    assert len(ref_node.children) == 3
    # We have an @ symbol, the reference name and in a child node,
    # the value string.
    ref = ref_node.children[1].text
    has_value = bool(ref_node.children[2].children)
    if has_value:
        value = parse_str(ref_node.children[2].children[0].children[1], charset)
    else:
        value = None
    return Ref(ref, value, has_value)

def parse_date(date_node):
    assert date_node.expr_name == 'date'
    # Date is in 3 parts, separated by hyphens.
    (ys, ms, ds) = date_node.text.split('-',3)
    return datetime.date(year=int(ys),
                        month=int(ms),
                        day=int(ds))

def parse_time(time_node):
    assert time_node.expr_name == 'time'
    # Date is in 3 parts, separated by hyphens.
    (hs, ms, ss) = time_node.text.split(':',3)
    # Split seconds into whole and fractional parts
    if '.' in ss:
        (whole_sec, frac_sec) = ss.split('.',1)
        sec = int(whole_sec)
        usec = int(frac_sec[:6].ljust(6,'0'))
    else:
        sec = int(ss)
        usec = 0
    return datetime.time(hour=int(hs),
                        minute=int(ms),
                        second=sec,
                        microsecond=usec)

def parse_date_time(date_time_node):
    assert date_time_node.expr_name == 'dateTime'
    assert len(date_time_node.children) == 2
    # Made up of parts: ISO8601 Date/Time, time zone label
    # (with proceeding space)
    isodt = iso8601.parse_date(date_time_node.children[0].text)
    tzname = date_time_node.children[1].text[1:]

    if (isodt.tzinfo is None) and bool(tzname): # pragma: no cover
        # This technically shouldn't happen according to Zinc specs
        return timezone(tzname).localise(isodt)
    elif bool(tzname):
        try:
            tz = timezone(tzname)
            return isodt.astimezone(tz)
        except: # pragma: no cover
            # Unlikely to occur, might do though if Project Haystack changes
            # its timezone list or if a system doesn't recognise a particular
            # timezone.
            return isodt    # Failed, leave alone
    else:
        return isodt
